#include "hogwash.h"
/******************************************
* hogwash.c
*
* a simple libpcap packet filter
* designed to be used as an inline filter.
******************************************/
#include "pass_rule.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdarg.h>

/*global variables*/
struct interface_rec	internal;
struct interface_rec	external;
char*					internal_name;
char*					external_name;

struct SaveRec*			int_head;
struct SaveRec*			ext_head;
struct SaveRec			int_recs[NUMBUFFS];
struct SaveRec			ext_recs[NUMBUFFS];
pthread_mutex_t			int_packet_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t			ext_packet_mutex = PTHREAD_MUTEX_INITIALIZER;

int						no_rules;
FILE* 				stream;
int				debuglevel;
int				loglevel;

/*function prototypes*/
void InitStructures();
void InitInterface(struct interface_rec* r);
void CloseInterfaces();
void ShowBanner(int args);
int GetIfrMTU(char *name) ;
void ProcessPacket(char *user, struct pcap_pkthdr * pkthdr, u_char * pkt);
int OpenPcap(struct interface_rec* r, char* interface_name);
void CloseRawSock(struct interface_rec *r);
void* MainLoop(void* v);
void ParseArgs(int argc, char* argv[]);
void CreateLock();
void DestroyLock();
int hogwashprint (int loglevel, const char* format,...);


/****************************************************
* The one and only main
****************************************************/
int main(int argc, char* argv[]){
	int_head=NULL;
	ext_head=NULL;

	/*testing*/
	hogwashprint(5,"binding signals\n");
	signal(SIGTERM, DestroyLock);
	signal(SIGINT,  DestroyLock);
	signal(SIGQUIT, DestroyLock);
	/*end testing*/

	ShowBanner(FALSE);
	InitStructures();
	ParseArgs(argc, argv);
	CreateLock();
#ifdef DEBUG
	debuglevel=5;
#endif
	InitInterface(&internal);
	InitInterface(&external);
	if (!no_rules) pass_rule_init(internal.pds, argc, argv);
	/*start up the handlers*/
	pthread_create(&external.thread, NULL, MainLoop, (void*)&external);
	MainLoop((void*)&internal);
	
	hogwashprint(5,"About to close interfaces\n");
	CloseInterfaces();
	pass_rule_free();
	DestroyLock();
	exit(0);
}

/****************************************************
* Creates a PID so we don't run 2
****************************************************/
void CreateLock(){
	int			fp;
	struct stat	s;
	int			pid;
	int			statv;
	char		Buff[1024];
	
	bzero(&s, sizeof(struct stat));
	statv=stat(PID_FILE, &s);
	if (statv!=0){
		fp = open(PID_FILE, O_CREAT|O_RDWR);
		if (!fp){
			hogwashprint(5,"Couldn't create PID file %s.\n",PID_FILE);
			exit(1);
		}
		pid=getpid();
		sprintf(Buff,"%i\n",pid);
		write(fp, Buff, strlen(Buff));
		close(fp);
		return;
	}
	
	hogwashprint(5,"PID file %s exists. Running two copies of hogwash may crash some switches.\n", PID_FILE);  
	exit(1);
	
}

/***************************************************
* Erase the PID 
***************************************************/
void DestroyLock(){
	CloseInterfaces();
	pass_rule_free();
	remove(PID_FILE);
	exit(0);
}

/****************************************************
* The one and only ParseArgs :)
****************************************************/
void ParseArgs(int argc, char* argv[]){
	int	i=0;
	
	if (argc==1) ShowBanner(TRUE);
	
	for (i=0; i<argc; i++){	
		if (strncasecmp("-i",argv[i],2)==0){
			if ( (i>=argc-1) || (argv[i+1][0]=="-"[0])){
				hogwashprint(5,"Syntax -i <internal interface>\n");
				exit(1);
			}
			snprintf(internal_name, 10, argv[i+1]);
			hogwashprint(5,"New Internal %s\n",internal_name);
			i++;
		}else if (strncasecmp("-e",argv[i],2)==0){
			if ( (i>=argc-1) || (argv[i+1][0]=="-"[0]) ){
				hogwashprint(5,"Syntax -e <external interface>\n");
				exit(1);
			}
			snprintf(external_name, 10, argv[i+1]);
			hogwashprint(5,"New External %s\n",external_name);
		}else if (strncasecmp("-n", argv[i],2)==0){
			no_rules=TRUE;
			hogwashprint(5,"******************************************\n");
			hogwashprint(5,"* Entering No Rules Mode\n");
			hogwashprint(5,"******************************************\n");
		}
		else if (strncasecmp("-h",argv[i],2)==0)
			{
			ShowBanner(TRUE);
			}

	}
}

/********************************************
* one for each thread
********************************************/
void* MainLoop(void* v){
	struct interface_rec*	r;
	
	r=(struct interface_rec*)v;
	
	pcap_loop(r->pds, -1, (pcap_handler) ProcessPacket, (char*)r->num);
	return NULL;
}

/***************************************
* print the banner so we know who we are
***************************************/
void ShowBanner(int args){
	printf("hogwash v0.01c by Jason Larsen\n");
	printf("using snort v1.7 by Martin Roesch\n\n");
	if (args){
		printf("Usage hogwash <options>\n");
		printf("  -c <rules file>\n");
		printf("  -i <internal interface>\n");
		printf("  -e <external interface>\n");
		printf("  -l <log directory>\n");
		printf("  -n (no rules mode)\n");
		printf("  -h (show this help message and exit)\n");
		exit(1);
	}
}

/**************************************
* set all variables to default values
**************************************/
void InitStructures(){
	no_rules=FALSE;
	
	bzero(int_recs, sizeof(struct SaveRec) * NUMBUFFS);
	bzero(int_recs, sizeof(struct SaveRec) * NUMBUFFS);
	
	stream=fopen (OUTPUT_FILE,"a");
	
	internal_name=(char*)malloc(10);
	sprintf(internal_name, "eth0");
	external_name=(char*)malloc(10);
	sprintf(external_name, "eth1");

	internal.num=0;
	internal.string=internal_name;
	internal.pds=NULL;
	internal.snaplen = SNAPLEN;
	internal.mtu=0;
	internal.datalink=0;
	
	external.num=1;
	external.string=external_name;	
	external.pds=NULL;
	external.snaplen = SNAPLEN;
	external.mtu=0;
	external.datalink=0;
}

/**************************************
* startup the libpcap interfaces
**************************************/
void InitInterface(struct interface_rec* r){
	OpenPcap(r, r->string);
	OpenRawSock(r);
}

/***************************************
* Shut everything down
***************************************/
void CloseInterfaces(){
	CloseRawSock(&internal);
	CloseRawSock(&external);
	fclose (stream);
}

/*******************************************************
 * open the libpcap devices.  Mostly stolen from snort.
 *******************************************************/
int OpenPcap(struct interface_rec* r, char* interface_name){
	char errorbuf[1024];
    struct bpf_program fcode;    
	
	printf("In OpenPcap\n");
	if (!interface_name){
		r->string = pcap_lookupdev(errorbuf);
		hogwashprint(5,"found interface %s\n", r->string);
	}else{
		r->string=interface_name;
		hogwashprint(5,"Using guess %s\n", r->string);
	}
	r->pds = pcap_open_live(r->string, SNAPLEN,
         PROMISC , READ_TIMEOUT, errorbuf);
	 /* lookup mtu */
    r->mtu = GetIfrMTU(r->string);
	hogwashprint(5,"New MTU: %i\n",r->mtu);
	if (r->mtu == -1){
		printf("Couldn't get MTU\n");
		return FALSE;
	}
	
	if(pcap_compile(r->pds, &fcode, NULL, 1, 0) < 0)
    {
        hogwashprint(3,"ERROR: OpenPcap() FSM compilation failed: \n\t%s\n",
                     pcap_geterr(r->pds));
        hogwashprint(3,"PCAP command: %s\n", NULL);
    }
	
    /* set the pcap filter */
    if(pcap_setfilter(r->pds, &fcode) < 0)
    {
        hogwashprint(3,"ERROR: OpenPcap() setfilter: \n\t%s\n",
                   pcap_geterr(r->pds));
    }
	
    /* get data link type */
    r->datalink = pcap_datalink(r->pds);

    if(r->datalink < 0)
    {
        hogwashprint(3,"ERROR: OpenPcap() datalink grab: \n\t%s\n",
                   pcap_geterr(r->pds));
    }
	
	return TRUE;
}

/****************************************************************************
 *
 * Function  : GetIfrMTU()
 * Purpose   : Get Interface MTU value
 * Arguments : interface name (string)
 * Returns   : MTU (or -1)
 *
 ****************************************************************************/


int GetIfrMTU(char *name) {
    
    int fd;
    struct ifreq ifr;
    int retval;


    retval = -1;
    
    fd = socket(AF_INET, SOCK_DGRAM, 0);
    if ( fd < 0) {
        hogwashprint(1,"socket");
        return -1;
    }

    strncpy(ifr.ifr_name, name, sizeof(ifr.ifr_name));

#ifndef OSF1
    if (ioctl(fd, SIOCGIFMTU, &ifr) == 0)
        retval = ifr.ifr_metric;
#else
    if (ioctl(fd, SIOCRIPMTU, &ifr) == 0)
        retval = ifr.ifr_metric;
#endif
    else
        printf("ioctl(SIOCGIFMTU)");    

    close(fd);
    
 return retval;
}


/***********************************************
* Push a packet onto the queue
***********************************************/
void PushPacketOld(char* packet, int len, int which, struct timeval* tv){
	struct SaveRec*	seek;
	struct SaveRec*	new_packet;
	struct SaveRec*	del;
	
	/*construct the new packet*/
	new_packet=(struct SaveRec*)malloc(sizeof(struct SaveRec));
	bzero(&new_packet->Buff[0], BUFFSIZE);
	memcpy(&new_packet->Buff[0], packet, len);
	memcpy(&new_packet->timestamp, tv, sizeof(struct timeval));
	new_packet->next = NULL;
	
	if (which==0){
		seek = int_head;
		if (!int_head){
			int_head=new_packet;
			return;
		}
	}else{
		seek=ext_head;
		if (!ext_head){
			ext_head=new_packet;
			return;
		}
	}
	

	/*expire old packets*/
	while (seek->next){
		if ((tv->tv_sec - seek->next->timestamp.tv_sec)>2){
			del = seek->next;
			seek->next=seek->next->next;
			free(del);
		}else{
			seek=seek->next;
		}
	}

	seek->next = new_packet;
}

/***********************************************
* Push a packet onto the queue
***********************************************/
void PushPacket(char* packet, int len, int which, struct timeval* tv){
	struct SaveRec*	seek;
	int				i;
	
	
	if (which==0){
		seek=&int_recs[0];
	}else{
		seek=&ext_recs[0];
	}
	
	/*find an open slot*/
	for (i=0;i<NUMBUFFS;i++){
		if (seek[i].len==0){
			memcpy(seek[i].Buff, packet, len);
			seek[i].len=len;
			memcpy(&seek[i].timestamp, tv, sizeof(struct timeval));
			return;
		}
		/*expire the buffer*/
		if ((tv->tv_sec-seek[i].timestamp.tv_sec) > BUFFTIMEOUT){
			seek[i].len=0;
			i--;
		}
	}
}


/*********************************************
* Pop a packet off the queue
* return FALSE if not found
*********************************************/
int PopPacketOld(char* packet, int len, int which){
	struct SaveRec*	del;
	struct SaveRec*	head;
	struct SaveRec* last;
	
	if (which==0){
		head=ext_head;
		if (!ext_head){
			return FALSE;
		}
	}else{
		head=int_head;
		if (!int_head){
			return FALSE;
		}
	}
	
	while (head){
		if (memcmp(&head->Buff[0], packet, len)==0){
			
			if (head==int_head){
				del=int_head;
				int_head=int_head->next;
				free(del);
				del=NULL;
				return TRUE;
			}
			
			if (head==ext_head){
				del=ext_head;
				ext_head=ext_head->next;
				free(del);
				del=NULL;
				return TRUE;
			}
			
			del=head;
			last->next=head->next;
			free(del);
			del=NULL;
			return TRUE;
		}else{
			last=head;
			head=head->next;
		}
	}
	
	return FALSE;
}


/*********************************************
* Pop a packet off the queue
* return FALSE if not found
*********************************************/
int PopPacket(char* packet, int len, int which){
	struct SaveRec* seek;
	int				i;
	
	struct SaveRec*	del;
	struct SaveRec*	head;
	struct SaveRec* last;
	
	if (which==0){
		seek=ext_recs;
	}else{
		seek=int_recs;
	}
	
	for (i=0;i<NUMBUFFS;i++){
		if (seek[i].len==len)
		if (memcmp(seek[i].Buff, packet, len)==0){
			/*found it.  delete and bail*/
			seek[i].len=0;
			return TRUE;
		}
	}
	
	return FALSE;
}


void PrintPacket(unsigned char* p, int len){
	int i;
	
	for(i=0;i<len; i++){
		hogwashprint(2,"%02x", p[i]);
	}
	hogwashprint(2,"\n");
}


/*******************************************
* Each packet hits here
*******************************************/
void ProcessPacket(char *user, struct pcap_pkthdr * pkthdr, u_char * pkt){
	int						which;
	int						c;
	struct interface_rec*	r;
	
	which=(int)user;

	/*see if this packet is a duplicate*/	
	if (which==0){
		/*came in on the internal interface*/
		r=&external;
		
		/*compare the packet for duplicates*/
		pthread_mutex_lock(&ext_packet_mutex);
		if (PopPacket(pkt, pkthdr->len, which)){
			/*we are seeing a packet we just sent*/
			pthread_mutex_unlock(&ext_packet_mutex);
			return;
		}
		pthread_mutex_unlock(&ext_packet_mutex);
	}else{
		/*came in on the external interface*/
		r=&internal;
		
		pthread_mutex_lock(&int_packet_mutex);
		if (PopPacket(pkt, pkthdr->len, which)){
			/*we are seeing a packet we just sent*/
			pthread_mutex_unlock(&int_packet_mutex);
			return;
		}		
		pthread_mutex_unlock(&int_packet_mutex);
	}
	
	/*see if we drop this packet*/
	if (!no_rules)
	switch (pass_rule(user, pkthdr, pkt)){
		case RESPONSE_DROP:
			return;
		case RESPONSE_SEND_ON:
			break;
		case RESPONSE_SEND_REPLY:
			/*swap interfaces to send it out the correct one*/
			which= !which;
			if (which==0){
				r=&external;
			}else{
				r=&internal;
			}
			break;
		default:
			printf("Invalid Resonse Code\n");
			return;
	}
	
	/*make a copy to check for duplicates*/
	if (which==0){
		pthread_mutex_lock(&int_packet_mutex);
		PushPacket(pkt, pkthdr->len, which, &pkthdr->ts);
		pthread_mutex_unlock(&int_packet_mutex);
	}else{
		pthread_mutex_lock(&ext_packet_mutex);
		PushPacket(pkt, pkthdr->len, which, &pkthdr->ts);
		pthread_mutex_unlock(&ext_packet_mutex);
	}
	
	/*send the packet out the other interface*/
	c = libnet_write_link_layer(r->network, r->string, pkt, pkthdr->len); 
    if (c < pkthdr->len) 
    { 
        libnet_error(LN_ERR_WARNING, "libnet_write_link_layer only wrote %d bytes\n", c); 
    } 

}


/******************************************************
* creates a socket for writing
******************************************************/
int OpenRawSock(struct interface_rec* r){
    char 					err_buf[LIBNET_ERRBUF_SIZE]; 
		
	r->max_packet=1600;  /*ethernet*/
   	
	if ((r->network = libnet_open_link_interface(r->string, err_buf)) == NULL) 
    { 
        libnet_error(LIBNET_ERR_FATAL, "libnet_open_link_interface: %s\n", err_buf); 
    } 
	
	/*malloc max packet size*/
	 if (libnet_init_packet(r->max_packet, &r->packet) == -1) 
    { 
        libnet_error(LIBNET_ERR_FATAL, "libnet_init_packet failed\n"); 
    } 
 
	return TRUE;
}

/***********************************************
* Shut down the libnet interface
***********************************************/
void CloseRawSock(struct interface_rec *r){
    /* 
     *  Shut down the interface. 
     */ 
    if (libnet_close_link_interface(r->network) == -1) 
    { 
        libnet_error(LN_ERR_WARNING, "libnet_close_link_interface couldn't close the interface"); 
    } 
  

    /* 
     *  Free packet memory. 
     */ 
    libnet_destroy_packet(&r->packet); 
}

/* hogwashprint: print to either stdout or FILE* stream, depending on the
difference between loglevel and debuglevel.  If loglevel is less than or
equal to debuglevel, we vprintf to stdout; if it's greater than, we
vfprintf to stream.  We return the number of characters output */

int hogwashprint (int loglevel,const char* format,...)
	{
	va_list varargs;
	int file_or_stdout;
	va_start (varargs, format);
	file_or_stdout=loglevel-debuglevel;
	if (file_or_stdout >= 0 )
		{
		vprintf (format, varargs);
		}
	else
		{
		vfprintf (stream, format, varargs);
		}
		va_end (varargs);
	return (sizeof (format));
	}
	
	

